#!/usr/bin/python
"""
Match Game
by Kris Schnee.

A playable piece of a puzzle game, using the basic game mechanic
of the game "Bejeweled" and numerous other puzzle games:
match 3+ objects of a kind in a row or column to make them vanish.
"""

##___________________________________
##          ~Header~
##___________________________________

__author__ = "Kris Schnee"
__version__ = "2008.2.17"
__license__ = "Public Domain"


##___________________________________
##      ~Imported Modules~
##___________________________________

## Standard
import random
import os

## Third-party
import pygame
from pygame.locals import *

## Mine
import ringtale

##___________________________________
##            ~Constants~
##___________________________________

TILE_KINDS = "rrggbbpc"
TILE_SIZE = 64

##___________________________________
##            ~Classes~
##___________________________________

##random.seed("Aqua Stream")

class Tile:
    def __init__(self,coords,kind):
        self.coords = coords
        self.screen_coords = [self.coords[0]*TILE_SIZE,self.coords[1]*TILE_SIZE]
        self.kind = kind

    def SetCoords(self,coords):
        self.coords = coords
        self.screen_coords = [self.coords[0]*TILE_SIZE,self.coords[1]*TILE_SIZE]




class MatchGameBoard:
    def __init__(self):
        self.board_size = (8,8)
        self.grid = []
        for y in range(self.board_size[1]):
            for x in range(self.board_size[0]):
                kind = self.RandomTileKind()
                self.grid.append(Tile((x,y),kind))

    def RandomTileKind(self):
        kind = TILE_KINDS[random.randint(0,len(TILE_KINDS)-1)]
        return kind

    def __repr__(self):
        text = ""
        for y in range(self.board_size[1]):
            for x in range(self.board_size[0]):
                text += self.grid[y*8+x].kind
            text += "\n"
        return text

    def FindMatches(self):
        """Follow the standard match-game rule."""
        matches = []
        ## Horizontal matches
        for y in range(self.board_size[1]):
            start_x = 0
            current_kind = None
            for x in range(self.board_size[0]+1):
                if x == self.board_size[1]:
                    this_kind = " "
                else:
                    this_kind = self.grid[y*self.board_size[1]+x].kind
                if this_kind == current_kind:
                    pass
                else:
                    ## Chain ended. How long is it?
                    length = x - start_x
                    if length >= 3 and current_kind != " ":
                        matches.append({"kind":current_kind,"length":length,"axis":"x","x":(start_x,x-1),"y":y})
                        print matches[-1]
                    start_x = x
                    current_kind = this_kind

        ## Vertical matches
        for x in range(self.board_size[0]):
            start_y = 0
            current_kind = None
            for y in range(self.board_size[1]+1):
                if y == self.board_size[1]:
                    this_kind = " "
                else:
                    this_kind = self.grid[y*self.board_size[1]+x].kind
                if this_kind == current_kind:
                    pass
                else:
                    ## Chain ended. How long is it?
                    length = y - start_y
                    if length >= 3 and current_kind != " ":
                        matches.append({"kind":current_kind,"length":length,"axis":"y","x":x,"y":(start_y,y-1)})
                    start_y = y
                    current_kind = this_kind

        return matches

    def ClearTiles(self,sets_cleared):
        for s in sets_cleared:
            if s["axis"] == "x":
                for x in range(s["x"][0],s["x"][1]+1):
                    self.grid[s["y"]*self.board_size[0]+x].kind = " "
            elif s["axis"] == "y":
                for y in range(s["y"][0],s["y"][1]+1):
                    self.grid[y*self.board_size[0]+s["x"]].kind = " "

    def MakeTilesFall(self):
        for n in range(self.board_size[1]):
            for y in range(self.board_size[1]-1):
                for x in range(self.board_size[0]):
                    below = self.grid[(y+1)*self.board_size[0]+x]
                    if below.kind == " ":
                        this_tile = self.grid[(y)*self.board_size[0]+x]
                        self.grid[(y+1)*self.board_size[0]+x] = this_tile
                        self.grid[y*self.board_size[0]+x] = below
                        this_tile.SetCoords((x,y+1))
                        below.SetCoords((x,y))

    def SwitchTiles(self,c1,c2):
        tile_1 = self.grid[c1[1]*self.board_size[0]+c1[0]]
        tile_2 = self.grid[c2[1]*self.board_size[0]+c2[0]]
        self.grid[c1[1]*self.board_size[0]+c1[0]] = tile_2
        self.grid[c2[1]*self.board_size[0]+c2[0]] = tile_1
        tile_2.SetCoords(c1)
        tile_1.SetCoords(c2)



class Game( ringtale.Framework ):
    def __init__(self):
        ringtale.Framework.__init__(self)
        self.board = MatchGameBoard()

        self.selected_tile = None

        self.backgrounds = {}
##        self.backgrounds["coon"] = pygame.image.load(os.path.join("graphics","bg_raccoon.png")).convert()
##        self.backgrounds["gold"] = pygame.image.load(os.path.join("graphics","bg_gold.png")).convert()
        self.backgrounds["coon"] = pygame.image.load(os.path.join("graphics","bg_raccoon.jpg")).convert()
        self.backgrounds["gold"] = pygame.image.load(os.path.join("graphics","bg_gold.jpg")).convert()

        self.current_background = self.backgrounds["coon"]

        self.tile_size = 64
        try:
            self.tiles = {}
            FILES = {" ":"empty","p":"pc_gem_pink","r":"pc_gem_red","g":"pc_gem_green","b":"pc_gem_blue","c":"pc_chest"}
            for f in FILES:
                tile = pygame.image.load(os.path.join("graphics",FILES[f]+".png"))
                tile.set_colorkey((255,0,255))
                self.tiles[f] = tile
        except:
            COLORS = {"r":(255,0,0),"g":(0,255,0),"b":(0,0,255),"y":(255,255,0)," ":(0,0,0)}

            self.tiles = {}
            for key in COLORS:
                s = pygame.surface.Surface((64,64))
                s.fill(COLORS[key])
                self.tiles[key] = s

        self.PushState("Main")

    def MainLogic(self):
        for event in pygame.event.get():
            if event.type == QUIT:
                self.PopState()
            elif event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.PopState()

            elif event.type == MOUSEBUTTONDOWN:
                tile_clicked = self.GetTileCoordsClicked(event.pos)
                if tile_clicked:
                    if self.selected_tile:
                        self.board.SwitchTiles(self.selected_tile,tile_clicked)
                        self.selected_tile = None
                    else:
                        self.selected_tile = tile_clicked

        matches = self.board.FindMatches()
        if matches:
            self.board.ClearTiles(matches)

        self.board.MakeTilesFall()

    def MainDraw(self):
        self.DrawBoard()
        if self.selected_tile:
            x, y = self.selected_tile
            pygame.draw.rect(screen,(0,255,255),(x*64,y*64,64,64),3)

    def GetTileCoordsClicked(self,screen_coords):
        absolute_x = screen_coords[0]
        absolute_y = screen_coords[1]
        tile = (absolute_x/self.tile_size,absolute_y/self.tile_size)
        if tile[0] >= 0 and tile[0] < self.board.board_size[0] and tile[1] >= 0 and tile[1] < self.board.board_size[1]:
            return tile
        else:
            return None ## Impossible location.

    def DrawBoard(self):
        screen.blit(self.current_background,(0,0))    
        for y in range(self.board.board_size[1]):
            for x in range(self.board.board_size[0]):
                tile = self.board.grid[y*self.board.board_size[0]+x]
                kind = tile.kind
                sx, sy = x*64, y*64
                try:
                    assert [sx, sy] == tile.screen_coords
                except:
                    raise "Tile "+str((x,y))+": "+str(((sx, sy), tile.screen_coords))

                sx, sy = tile.screen_coords
                screen.blit(self.tiles[kind],(sx,sy))

        pygame.display.update()


##___________________________________
##            ~Functions~
##___________________________________


##__________________________________
##            ~Autorun~
##__________________________________

if __name__ == "__main__":

    screen = pygame.display.set_mode((512,512))

    g = Game()
    g.MainLoop()

